// RenderEngine class header
#pragma once

#include "client/Window.h"
#include "nf/config.h"
#include "CommandPool.h"
#include "VideoMemoryAllocator.h"
#include "Buffer.h"
#include "Image.h"

namespace nf::client::render {
    class RenderEngine {
    public:
        RenderEngine(std::shared_ptr<Window> window, DisplayConfig display);

        void doFrame();
        void setDisplay(DisplayConfig config);

        ~RenderEngine();
    private:
        void createInstance();
        void createSurface();
        void pickPhysicalDevice();
        void createDevice();
        void createExecutionObjects();

        void createSwapchain();

        void createOutputRenderPass();
        void createOutputPipeline();

        void createSwapchainFramebuffers();
        void createBuffers();
        void createImage();
        void createDescriptorSet();

        void recreateSwapchain();
        void destroySwapchain();

        void waitIdle();

        std::shared_ptr<Window> m_window;
        DisplayConfig m_display;

        // Global state
        VkInstance m_instance;
        VkSurfaceKHR m_surface;
        VkPhysicalDevice m_physicalDevice;
        uint32_t m_queueFIGraphics, m_queueFIPresent;
        VkDevice m_device;
        VkQueue m_queueGraphics, m_queuePresent;

        // Execution objects
        std::unique_ptr<CommandPool> m_commandPool;
        VkCommandBuffer m_commandBuffer;
        VkSemaphore m_semaphoreImageAvailable;
        VkSemaphore m_semaphoreRenderingDone;
        VkFence m_fenceFrameInFlight;

        // Swapchain
        VkSwapchainKHR m_swapchain;
        VkFormat m_swapchainImageFormat;
        VkExtent2D m_swapchainExtent;
        std::vector<VkImage> m_swapchainImages;
        std::vector<VkImageView> m_swapchainImageViews;

        // RenderPass and Pipeline
        VkRenderPass m_renderPassOutput;
        VkDescriptorSetLayout m_pipelineOutputDescriptorSetLayout;
        VkPipelineLayout m_pipelineOutputLayout;
        VkDescriptorPool m_pipelineOutputDescriptorPool;
        VkPipeline m_pipelineOutput;

        // Temporary objects
        std::unique_ptr<VideoMemoryAllocator> m_allocator;
        std::unique_ptr<Image> m_imageDepth;
        std::vector<VkFramebuffer> m_swapchainFramebuffers;
        std::unique_ptr<Buffer> m_bufferVertex;
        std::unique_ptr<Buffer> m_bufferIndex;
        std::unique_ptr<Buffer> m_bufferUniformMVP;
        std::unique_ptr<Image> m_imageTest;
        VkSampler m_sampler;
        VkDescriptorSet m_pipelineOutputDescriptorSet;
    };
}
